#######################################################################################
###### This package handles the interaction with the ERS switch via SNMP     ##########
#######################################################################################
package ERS_MIB;
our $VERSION = '0.02';

use strict;
use Net::SNMP qw(:snmp);
use Sort::Key::OID qw(oidsort);
use MIBstuff 0.01;
use ERS_data 0.01;

#################################### init Variables ###################################
my $maxBulkRepition = 30;
my $retries = 1;
my $timeout = 2;

#######################################################################################
# new
sub new {
	my $class = shift;
	my $comunity = shift;
	my $debug = shift || 0;
	my $self = {};
	
	bless($self, $class);
	
	$self->{_name}		= '';
	$self->{'debug'}	= $debug;
	$self->{_comunity}	= $comunity;
	$self->{_ERRMSG}	= '';
	$self->{'oids'}		= [];
	
	return $self;
}

#######################################################################################
#DESTROY();
sub DESTROY {
	my $self = shift;
	
}

#######################################################################################
#getErrorMsg();
sub getErrorMsg {
	my $self = shift;
	
	return $self->{_ERRMSG};
}

#######################################################################################
#clearErrorMsg();
sub clearErrorMsg {
	my $self = shift;
	
	$self->{_ERRMSG} = '';
}

#######################################################################################
#performSNMP(job, action, Host, MIBname, callback)
sub performSNMP {
	my $self 	= shift;
	my $action 	= shift;
	my $callBack	= shift;
	my $refA_hosts	= shift;
	my $refA_MIBs	= shift;
	my $refO_data	= shift;
	my $maxRepetitions = shift || 10;						# only for get_bulk_request
	$callBack .= '_cb';
	
	my %params	= (	-varbindlist	=> $refA_MIBs,
				-callback	=> [ \&$callBack, $refA_hosts, $refO_data, $action, $refA_MIBs->[0]],
			);
	
	if($action eq 'walk_request'){
		$action = 'get_bulk_request';
	}
	
	if($action eq 'get_bulk_request'){
		$params{'-maxrepetitions'} = $maxRepetitions;
	}
	
	print "[**.**.**.**] DEBUG: execute SNMP $action ".@{$refA_MIBs}."/$maxRepetitions\n" if $self->{'debug'};
	
	foreach my $ip (@{$refA_hosts}){
#		print "$ip => $callBack\n";
		my ($session, $error) = Net::SNMP->session(	-version	=> 'snmpv2c',
								-hostname   	=> $ip,
								-community   	=> $self->{_comunity},
#								-translate	=> [-octetstring => 0],
								-retries	=> $retries,
								-timeout	=> $timeout,
								-nonblocking	=> 1,
							);

		if (!defined($session)) {
			print "ERROR: SNMP initialisation\n";
			$self->{_ERRMSG} = $error;
			die;
		}

		my $result = $session->$action(%params);
		
		if (!defined($result)) {
			printf "ERROR: Failed to queue get request for host '%s': %s.\n",
			$session->hostname(), $session->error();
			$self->{_ERRMSG} = $error;
			$session->close;
#			return $result;
		}
	}
	
	print "[**.**.**.**] DEBUG: waiting for response\n" if $self->{'debug'};
	Net::SNMP::snmp_dispatcher();
	
	print "[**.**.**.**] DEBUG: all responses collected\n" if $self->{'debug'};
	removeBadIp($refA_hosts);
}

#######################################################################################
#determineOID()
sub determineOID {
	my $self	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my @MIBs	= ($sysDescr, $sysObjectID, $sysName);
	
	$self->performSNMP('get_request', 'determineOID', $refA_hosts, \@MIBs, $refO_data);
}

#######################################################################################
#determineOID_cb()
sub determineOID_cb {
	my $session	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my $result	= $session->var_bind_list();
	my $ip		= $session->hostname;
	
  	if($session->{'_error'}){
#		print "ERROR: determineOID '".$session->{'_error'}."'\n";
  		markBadIp($refA_hosts, $ip);
	}else{
		$$refO_data{$ip}->setValue('OID', $result->{$sysObjectID});
		$$refO_data{$ip}->setValue('Name', $result->{$sysName});
		$$refO_data{$ip}->setValue('Description', $result->{$sysDescr});
		$result->{$sysDescr} =~ /SW\:v(\d.\d.\d.\d*) /;
		$$refO_data{$ip}->setValue('Software', $1);
  	}
}

#######################################################################################
#getUnitCount()
sub getUnitCount {
	my $self	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my @MIBs	= ($s5ChasComDescr);
	
	$self->performSNMP('get_bulk_request', 'getUnitCount', $refA_hosts, \@MIBs, $refO_data);
	
	foreach my $ip (@{$refA_hosts}){
#		print "getUnitCount $ip => ".$$refO_data{$ip}->getValue('UnitCount')."\n";
	}
}

#######################################################################################
#getUnitCount_cb()
sub getUnitCount_cb {
	my $session	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my $result	= $session->var_bind_list();
	my $ip		= $session->hostname;
	
	if($session->{'_error'}){
		print "ERROR: getUnitCount '".$session->{'_error'}."'\n";
  		markBadIp($refA_hosts, $ip);
	}else{
		my $count = 0;
		foreach my $mib (keys %{$result}){
			next if $mib !~ /$s5ChasComDescr/;
			$count++
		}
		$$refO_data{$ip}->setValue('UnitCount', $count);
	}
}

#######################################################################################
#getPortCount()
sub getPortCount {
	my $self	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my @MIBs	= ($ifNumber);
	
	$self->performSNMP('get_request', 'getPortCount', $refA_hosts, \@MIBs, $refO_data);
	
#	foreach my $ip (@{$refA_hosts}){
#		print "getPortCount $ip => ".$$refO_data{$ip}->getValue('PortCount')."\n";
#	}
}

#######################################################################################
#getPortCount_cb()
sub getPortCount_cb {
	my $session	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my $result	= $session->var_bind_list();
	my $ip		= $session->hostname;
	
	if($session->{'_error'}){
		print "ERROR: getPortCount '".$session->{'_error'}."'\n";
  		markBadIp($refA_hosts, $ip);
	}else{
		$$refO_data{$ip}->setValue('PortCount', $result->{$ifNumber});
	}
}

#######################################################################################
#userQuery()
sub userQuery {
	my $self	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my $refH_query	= shift;
	
	if (ref($refH_query->{'oid'}) eq 'ARRAY'){
		foreach my $oid (@{$refH_query->{'oid'}}){
			if($oid !~ /^(\d+\.)+\d+$/){
				print "ERROR: oid '$oid' is not valide, drop request\n";
				return;
			}
		}
	}elsif($refH_query->{'oid'} !~ /^(\d+\.)+\d+$/){
		print "ERROR: oid '$refH_query->{'oid'}' is not valide, drop request\n";
		return;
	}
	
	if($refH_query->{'methode'} =~ /^(get_request|get_next_request|get_bulk_request|walk_request)$/){
		if($refH_query->{'methode'} =~ /(get_bulk_request|walk_request)/){
			if($refH_query->{'repetitions'} =~ /^\d+$/){
				if($refH_query->{'repetitions'} > $maxBulkRepition){
					$refH_query->{'repetitions'} = $maxBulkRepition;
					print "[**.**.**.**] WARN:  repetition counter '$refH_query->{'repetitions'}'  is to high, set to $maxBulkRepition\n";
				}
			}
		}elsif(defined($refH_query->{'repetitions'})){
			$refH_query->{'repetitions'} = undef;
			print "[**.**.**.**] WARN:  ignore repetition counter because isn't applicable for methode $refH_query->{'methode'}\n";
		}
		
		if($refH_query->{'methode'} !~ /walk_request/){
			my @MIBs;
			if (ref($refH_query->{'oid'}) eq 'ARRAY'){
				@MIBs = @{$refH_query->{'oid'}};
			}else{
				push(@MIBs, $refH_query->{'oid'});
			}
			foreach my $host (@{$refA_hosts}){
				$$refO_data{$host}->setValue('calledOid', $refH_query->{'oid'});
			}
			$self->performSNMP($refH_query->{'methode'}, 'userQuery', $refA_hosts, \@MIBs, $refO_data, $refH_query->{'repetitions'});
		}else{
			if (ref($refH_query->{'oid'}) eq 'ARRAY'){
				print "[**.**.**.**] WARN:  ignore methode $refH_query->{'methode'}, because is more then one OID defined\n";
				return;
			}
			
			resetLastIndex($refA_hosts, $refO_data);
			my @hosts = @{$refA_hosts};
			foreach my $host (@hosts){
				$$refO_data{$host}->setValue('calledOid', $refH_query->{'oid'});
			}
			
			my $maxRepition = $refH_query->{'repetitions'};
			my $lastIndex	= undef;
			do{
				my @MIBs;
				if(defined($lastIndex)){
					push(@MIBs, $refH_query->{'oid'}.'.'.$lastIndex);
				}else{
					push(@MIBs, $refH_query->{'oid'});
				}
				print "[**.**.**.**] DEBUG: SNMP-WALK with OID $MIBs[0]\n" if $self->{'debug'};
				$self->performSNMP('walk_request', 'userQuery', \@hosts, \@MIBs, $refO_data, $maxRepition);
				
				my $index = undef;
				for(my $count = 0; $count < @hosts; $count++){
					print "[$hosts[$count]] DEBUG: " if $self->{'debug'};
					$lastIndex = $$refO_data{$hosts[$count]}->getValue('lastIndex');
					$lastIndex =~ s/\.(\d+)$/./;
					$index = $1;
					if($index >= 1000000 or $lastIndex >= 1000000){
						print "[$hosts[$count]] DEBUG: removed form list\n" if $self->{'debug'};
						splice(@hosts, $count, 1);
						$count--;
					}else{
						print "[$hosts[$count]] DEBUG: stay on list $lastIndex $index\n" if $self->{'debug'};
					}
				}
				
				if(defined($index)){
					$lastIndex .= ($index + $maxRepition - 1);
				}else{
					$lastIndex = $lastIndex + $maxRepition - 1;
				}
				print "[**.**.**.**] DEBUG: next SNMP-WALK run use index $lastIndex\n" if $self->{'debug'};
			}while(@hosts);
		}
	}else{
		print "[**.**.**.**] ERROR: methode '$refH_query->{'methode'}' is not a leagal SNMP methode, ignore query\n";
	}
}

#######################################################################################
#userQuery_cb()
sub userQuery_cb {
	my $session	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my $action	= shift;
	my $oid		= shift;
	my $result	= $session->var_bind_list();
	my $ip		= $session->hostname;
	my $lastIndex	= 0;
	my $pattern	= '^'.$$refO_data{$ip}->getValue('calledOid');
	
  	if($session->{'_error'}){
  		markBadIp($refA_hosts, $ip);
	}else{
		if($action ne 'walk_request'){
			foreach my $mib (oidsort keys %{$result}){
				$mib =~ /$pattern/;
				print "[$ip] response $' = '$result->{$mib}'\n";
			}
		}else{
			foreach my $mib (oidsort keys %{$result}){
				if($mib !~ /$pattern/){
					$lastIndex = 1000000;
					last;
				}
				
				$mib =~ /$pattern/;
				my $res = $';
				
				if($res){
					$lastIndex = $res if $lastIndex < $res;
					$lastIndex =~ s/^\.//;
				}else{
					$lastIndex = 1000000 if !$lastIndex;
				}
				
				print "[$ip] response $res = '$result->{$mib}'\n";
			}
			$$refO_data{$ip}->setValue('lastIndex', $lastIndex);
		}
  	}
}

#######################################################################################
#resetLastIndex()
sub resetLastIndex {
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	
	foreach my $ip (@{$refA_hosts}){
		$$refO_data{$ip}->setValue('lastIndex', 0);
	}
}

#######################################################################################
#markBadIp()
sub markBadIp {
	my $refA_hosts	= shift;
	my $badIp = shift;
	
	for(my $count = 0; $count < @{$refA_hosts}; $count++){
		if($$refA_hosts[$count] == $badIp){
			$$refA_hosts[$count] = 0;
			print "[$badIp] WARN:  bad IP address will be removed form processing\n";
			return;
		}
	}
}

#######################################################################################
#removeBadIp()
sub removeBadIp {
	my $refA_hosts	= shift;
	
	for(my $index = 0; $index < @{$refA_hosts}; $index++){
		if(!$$refA_hosts[$index]){
			splice(@{$refA_hosts}, $index, 1);
			$index--;
		}
	}
}

1;
